import json
from collections import Counter
from typing import List, Optional, Dict, Tuple

from presidio_evaluator.evaluation import ModelError


class EvaluationResult:
    def __init__(
        self,
        results: Counter,
        model_errors: Optional[List[ModelError]] = None,
        text: Optional[str] = None,
        pii_recall: Optional[float] = None,
        pii_precision: Optional[float] = None,
        pii_f: Optional[float] = None,
        n: Optional[int] = None,
        entity_recall_dict: Optional[Dict[str, float]] = None,
        entity_precision_dict: Optional[Dict[str, float]] = None,
        n_dict: Optional[Dict[str, int]] = None,
    ):
        """
        Holds the output of a comparison between ground truth and predicted
        :param results: List of objects of type Counter
        with structure {(actual, predicted) : count}
        :param model_errors: List of specific model errors for further inspection
        :param text: sample's full text (if used for one sample)
        :param pii_recall: Recall for all entities (PII or not)
        :param pii_precision: Precision for all entities (PII or not)
        :param pii_f: F measure for all entities (PII or not)
        :param n: Number of total entity tokens
        :param entity_recall_dict: Recall per entity
        :param entity_precision_dict: Precision per entity
        :param n_dict: Number of tokens per entity
        """

        self.results = results
        self.model_errors = model_errors
        self.text = text

        self.pii_recall = pii_recall
        self.pii_precision = pii_precision
        self.pii_f = pii_f
        self.n = n
        self.entity_recall_dict = entity_recall_dict
        self.entity_precision_dict = entity_precision_dict
        self.n_dict = n_dict

    def __str__(self):
        return_str = ""
        if not self.entity_precision_dict or not self.entity_recall_dict:
            return json.dumps(self.results)

        entities = self.n_dict.keys()

        row_format = "{:>20}{:>20.2%}{:>20.2%}{:>20}"
        header_format = "{:>20}" * 4
        return_str += str(
            header_format.format(
                *("Entity", "Precision", "Recall", "Number of samples")
            )
        )
        for entity in entities:
            return_str += "\n" + row_format.format(
                entity,
                self.entity_precision_dict[entity],
                self.entity_recall_dict[entity],
                self.n_dict[entity],
            )

        # add PII values
        return_str += "\n" + row_format.format(
            "PII",
            self.pii_precision,
            self.pii_recall,
            self.n,
        )

        return_str += f"\nPII F measure: {self.pii_f:.2%}"
        return return_str

    def __repr__(self):
        return f"stats={self.results}"

    def to_log(self):
        metrics_dict = {
            "pii_f": self.pii_f,
        }
        if self.entity_precision_dict:
            metrics_dict.update(
                {
                    f"{ent}_precision": v
                    for (ent, v) in self.entity_precision_dict.items()
                }
            )
        if self.entity_recall_dict:
            metrics_dict.update(
                {f"{ent}_recall": v for (ent, v) in self.entity_recall_dict.items()}
            )
        if self.n:
            metrics_dict.update(self.n_dict)
        return metrics_dict

    def to_confusion_matrix(self) -> Tuple[List[str], List[List[int]]]:
        entities = sorted(list(set(self.n_dict.keys()).union("O")))
        confusion_matrix = [[0] * len(entities) for _ in range(len(entities))]
        for i, actual in enumerate(entities):
            for j, predicted in enumerate(entities):
                confusion_matrix[i][j] = self.results[(actual, predicted)]

        return entities, confusion_matrix



import copy
from abc import ABC, abstractmethod
from typing import List, Dict, Optional

from presidio_evaluator import InputSample, io_to_scheme


class BaseModel(ABC):
    def __init__(
        self,
        labeling_scheme: str = "BIO",
        entities_to_keep: List[str] = None,
        entity_mapping: Optional[Dict[str, str]] = None,
        verbose: bool = False,
    ):

        """
        Abstract class for evaluating NER models and others
        :param entities_to_keep: Which entities should be evaluated? All other
        entities are ignored. If None, none are filtered
        :param labeling_scheme: Used to translate (if needed)
        the prediction to a specific scheme (IO, BIO/IOB, BILUO)
        :param entity_mapping: Dictionary for mapping this model's input and output with the expected.
        Keys should be the input entity types (from the input dataset),
        values should be the model's supported entity types.
        :param verbose: Whether to print more debug info


        """
        self.entities = entities_to_keep
        self.labeling_scheme = labeling_scheme
        self.entity_mapping = entity_mapping
        self.verbose = verbose

    @abstractmethod
    def predict(self, sample: InputSample) -> List[str]:
        """
        Abstract. Returns the predicted tokens/spans from the evaluated model
        :param sample: Sample to be evaluated
        :return: List of tags in self.labeling_scheme format
        """
        pass

    def align_entity_types(self, sample: InputSample) -> None:
        """
        Translates the sample's tags to the ones requested by the model
        :param sample: Input sample
        :return: None
        """
        if self.entity_mapping:
            sample.translate_input_sample_tags(dictionary=self.entity_mapping)

    def align_prediction_types(
        self, tags: List[str], ignore_unknown: bool = True
    ) -> List[str]:
        """
        Turns the model's output from the model tags to the input tags.
        :param tags: List of tags (entity names in IO or "O")
        :param ignore_unknown: True if entity types not in entity_mapping should be translated to "O"
        :return: New tags
        """
        if not self.entity_mapping:
            return tags

        inverse_mapping = {v: k for k, v in self.entity_mapping.items()}
        new_tags = [
            InputSample.translate_tag(
                tag, dictionary=inverse_mapping, ignore_unknown=ignore_unknown
            )
            for tag in tags
        ]
        return new_tags

    def filter_tags_in_supported_entities(self, tags: List[str]) -> List[str]:
        """
        Replaces tags of unwanted entities with O.
        :param tags: Lits of tags
        :return: List of tags where tags not in self.entities are considered "O"
        """
        if not self.entities:
            return tags
        return [tag if self._tag_in_entities(tag) else "O" for tag in tags]

    def to_scheme(self, tags: List[str]):
        """
        Translates IO tags to BIO/BILUO based on the input labeling_scheme
        :param tags: Current tags in IO
        :return: Tags in labeling scheme
        """

        io_tags = [self._to_io(tag) for tag in tags]

        return io_to_scheme(io_tags=io_tags, scheme=self.labeling_scheme)

    def _ignore_unwanted_entities(
        self, dataset: List[InputSample]
    ) -> List[InputSample]:
        """
        Copy dataset and turn non-requested entity types into "O"
        :param dataset: Input dataset
        :return: Copy of dataset with requested entity types and "O" otherwise
        """
        entities_in_dataset = set()
        for sample in dataset:
            entities_in_dataset.update(set([span.entity_type for span in sample.spans]))
        entities_in_dataset.add("O")

        entities_to_keep = set(self.entities).intersection(entities_in_dataset)
        entities_to_ignore = entities_in_dataset.difference(self.entities)
        self.entity_mapping = {v: "O" for v in entities_to_ignore}
        self.entity_mapping.update({v: v for v in entities_to_keep})

        dataset = copy.copy(dataset)

        [self.align_entity_types(sample) for sample in dataset]
        return dataset

    @staticmethod
    def _to_io(tag):
        if "-" in tag:
            return tag[2:]
        return tag

    def to_log(self) -> Dict:
        """
        Returns a dictionary of parameters for logging purposes.
        :return:
        """
        return {
            "labeling_scheme": self.labeling_scheme,
            "entities_to_keep": self.entities,
        }

    def _tag_in_entities(self, tag: str):
        if not self.entities:
            return True

        if tag == "O":
            return True

        if tag[1] == "-":  # BIO/BILUO
            return tag[2:] in self.entities
        else:  # IO
            return tag in self.entities



from collections import defaultdict
import random
import numpy as np
from typing import List, Dict
import json

from presidio_evaluator import InputSample


def split_dataset(dataset: List[InputSample], ratios):
    """
    Splits a provided dataset into n groups, by the template_id attribute
    :param dataset: List of InputSamples to be splitted
    :param ratios:  list of percentages. The len of the list would be the len of the splits returned,
    e.g. [0.7,0.2,0.1] for train, test, validation
    """
    splits = []
    remaining_dataset = dataset
    remaining_ratio = 1.0

    if sum(ratios) > 1 or sum(ratios) < 0.999:
        raise ValueError("Ratios should sum to 1 and be in (0,1]")

    for ratio in ratios:
        if 1 >= ratio > 0:
            first_templates, second_templates = split_by_template(
                remaining_dataset, ratio / remaining_ratio
            )
            first_split = get_samples_by_pattern(remaining_dataset, first_templates)
            second_split = get_samples_by_pattern(remaining_dataset, second_templates)
            splits.append(first_split)
            remaining_dataset = second_split
            remaining_ratio -= ratio
        else:
            raise ValueError("Ratio needs to be in (0,1]")

    return tuple(splits)


def group_by_template(dataset: List[InputSample]) -> Dict[str, List[InputSample]]:
    """
    Creates a dict of key = template ID and value = List[InputSamples] for this template id
    """
    samples_pattern_tup = [(sample.template_id, sample) for sample in dataset]

    group_by_template = defaultdict(list)
    for sample in samples_pattern_tup:
        group_by_template[sample[0]].append(sample[1])

    return group_by_template


def split_by_template(input_samples: List[InputSample], train_pct: float = 0.7):
    """
    Splits a daset of type List[InputSample] into a tuple of train template IDs and test template IDs
    """
    samples_grpd = group_by_template(input_samples)

    templates = np.array(list(samples_grpd.keys()))
    train_ind = set(
        random.sample(range(len(templates)), round(train_pct * len(templates)))
    )

    test_ind = set(range(len(templates))) - train_ind

    return templates[list(train_ind)], templates[list(test_ind)]


def get_samples_by_pattern(input_samples, patterns_list):
    samples_grpd = group_by_template(input_samples)
    dataset = []
    for pattern in patterns_list:
        dataset.extend(samples_grpd[pattern])
    random.shuffle(dataset)

    return dataset


def save_to_json(samples, output_file):
    examples_dict = [example.to_dict() for example in samples]

    with open("{}".format(output_file), "w+", encoding="utf-8") as f:
        json.dump(examples_dict, f, ensure_ascii=False, indent=4)



"""E2E scoring pipelines for the different models"""

from typing import List, Optional

from presidio_analyzer import EntityRecognizer
from presidio_analyzer.nlp_engine import SpacyNlpEngine

from presidio_evaluator import InputSample
from presidio_evaluator.evaluation import EvaluationResult, Evaluator
from presidio_evaluator.models import (
    PresidioRecognizerWrapper,
    PresidioAnalyzerWrapper,
    BaseModel,
)


def score_model(
    model: BaseModel,
    entities_to_keep: List[str],
    input_samples: List[InputSample],
    verbose: bool = False,
    beta: float = 2.5,
) -> EvaluationResult:
    """
    Run data through a model and gather results and stats
    """

    print("Evaluating samples")

    evaluator = Evaluator(model=model, entities_to_keep=entities_to_keep)
    evaluated_samples = evaluator.evaluate_all(input_samples)

    print("Estimating metrics")
    evaluation_result = evaluator.calculate_score(
        evaluation_results=evaluated_samples, beta=beta
    )
    precision = evaluation_result.pii_precision
    recall = evaluation_result.pii_recall
    entity_recall = evaluation_result.entity_recall_dict
    entity_precision = evaluation_result.entity_precision_dict
    f = evaluation_result.pii_f
    errors = evaluation_result.model_errors
    #
    print(f"precision: {precision}")
    print(f"Recall: {recall}")
    print(f"F {beta}: {f}")
    print(f"Precision per entity: {entity_precision}")
    print(f"Recall per entity: {entity_recall}")

    if verbose:

        false_negatives = [
            str(mistake) for mistake in errors if mistake.error_type == "FN"
        ]
        false_positives = [
            str(mistake) for mistake in errors if mistake.error_type == "FP"
        ]
        other_mistakes = [
            str(mistake) for mistake in errors if mistake.error_type not in ["FN", "FP"]
        ]

        print("False negatives: ")
        print("\n".join(false_negatives))
        print("\n******************\n")

        print("False positives: ")
        print("\n".join(false_positives))
        print("\n******************\n")

        print("Other mistakes: ")
        print("\n".join(other_mistakes))

    return evaluation_result


def score_presidio_recognizer(
    recognizer: EntityRecognizer,
    entities_to_keep: List[str],
    input_samples: Optional[List[InputSample]] = None,
    labeling_scheme: str = "BILUO",
    with_nlp_artifacts: bool = False,
    verbose: bool = False,
) -> EvaluationResult:
    """
    Run data through one EntityRecognizer and gather results and stats
    """

    if not input_samples:
        print("Reading dataset")
        input_samples = InputSample.read_dataset_json("../../data/synth_dataset_v2.json")
    else:
        input_samples = list(input_samples)

    print("Preparing dataset by aligning entity names to Presidio's entity names")

    updated_samples = Evaluator.align_entity_types(
        input_samples, entities_mapping=PresidioAnalyzerWrapper.presidio_entities_map
    )

    model = PresidioRecognizerWrapper(
        recognizer=recognizer,
        entities_to_keep=entities_to_keep,
        labeling_scheme=labeling_scheme,
        nlp_engine=SpacyNlpEngine(),
        with_nlp_artifacts=with_nlp_artifacts,
    )
    return score_model(
        model=model,
        entities_to_keep=entities_to_keep,
        input_samples=updated_samples,
        verbose=verbose,
    )




from typing import List

from presidio_analyzer import RecognizerResult
from presidio_anonymizer import AnonymizerEngine

from presidio_evaluator.data_generator import PresidioDataGenerator


class PresidioPseudonymization(PresidioDataGenerator):
    def __init__(self, map_to_presidio_entities: bool = True, **kwargs):
        """
        Create pseudoanoymized data using Presidio by identifying real entities
        and replacing them with fake ones.
        :param kwargs: kwargs for PresidioDataGenerator
        :param map_to_presidio_entities:
        Whether to creating a mapping between Faker's providers and Presidio's entities
        """

        super().__init__(**kwargs)
        if map_to_presidio_entities:
            self.add_provider_alias("name", "PERSON")
            self.add_provider_alias("ipv4", "IP_ADDRESS")
            self.add_provider_alias("company", "ORGANIZATION")
            self.add_provider_alias("country", "LOCATION")
            self.add_provider_alias("credit_card_number", "CREDIT_CARD")
            self.add_provider_alias("iban", "IBAN_CODE")
            self.add_provider_alias("phone_number", "PHONE_NUMBER")
            self.add_provider_alias("url", "DOMAIN_NAME")
            self.add_provider_alias("ssn", "US_SSN")
            self.add_provider_alias("email", "EMAIL_ADDRESS")
            self.add_provider_alias("date_time", "DATE_TIME")

    def pseudonymize(
        self,
        original_text: str,
        presidio_response: List[RecognizerResult],
        count: int,
    ):
        """

        :param original_text: str containing the original text
        :param presidio_response: list of results from Presidio, to be used to know where entities are
        :param count: number of perturbations to return
        :return: List[str] with fake perturbations of original text
        """

        presidio_response = sorted(presidio_response, key=lambda resp: resp.start)

        anonymizer_engine = AnonymizerEngine()
        anonymized_result = anonymizer_engine.anonymize(
            text=original_text, analyzer_results=presidio_response
        )

        templated_text = anonymized_result.text
        templated_text = templated_text.replace(">", "}}").replace("<", "{{")
        fake_texts = [self.parse(templated_text, add_spans=False) for _ in range(count)]
        return fake_texts



from collections import Counter
from typing import List, Optional, Dict

import numpy as np
from tqdm import tqdm

from presidio_evaluator import InputSample
from presidio_evaluator.evaluation import EvaluationResult, ModelError
from presidio_evaluator.models import BaseModel


class Evaluator:
    def __init__(
        self,
        model: BaseModel,
        verbose: bool = False,
        compare_by_io=True,
        entities_to_keep: Optional[List[str]] = None,
    ):
        """
        Evaluate a PII detection model or a Presidio analyzer / recognizer

        :param model: Instance of a fitted model (of base type BaseModel)
        :param compare_by_io: True if comparison should be done on the entity
        level and not the sub-entity level
        :param entities_to_keep: List of entity names to focus the evaluator on (and ignore the rest).
        Default is None = all entities. If the provided model has a list of entities to keep,
        this list would be used for evaluation.
        """
        self.model = model
        self.verbose = verbose
        self.compare_by_io = compare_by_io
        self.entities_to_keep = entities_to_keep
        if self.entities_to_keep is None and self.model.entities:
            self.entities_to_keep = self.model.entities

    def compare(self, input_sample: InputSample, prediction: List[str]):

        """
        Compares ground truth tags (annotation) and predicted (prediction)
        :param input_sample: input sample containing list of tags with scheme
        :param prediction: predicted value for each token
        self.labeling_scheme

        """
        annotation = input_sample.tags
        tokens = input_sample.tokens

        if len(annotation) != len(prediction):
            print(
                "Annotation and prediction do not have the"
                "same length. Sample={}".format(input_sample)
            )
            return Counter(), []

        results = Counter()
        mistakes = []

        new_annotation = annotation.copy()

        if self.compare_by_io:
            new_annotation = self._to_io(new_annotation)
            prediction = self._to_io(prediction)

        # Ignore annotations that aren't in the list of
        # requested entities.
        if self.entities_to_keep:
            prediction = self._adjust_per_entities(prediction)
            new_annotation = self._adjust_per_entities(new_annotation)
        for i in range(0, len(new_annotation)):
            results[(new_annotation[i], prediction[i])] += 1

            if self.verbose:
                print("Annotation:", new_annotation[i])
                print("Prediction:", prediction[i])
                print(results)

            # check if there was an error
            is_error = new_annotation[i] != prediction[i]
            if is_error:
                if prediction[i] == "O":
                    mistakes.append(
                        ModelError(
                            error_type="FN",
                            annotation=new_annotation[i],
                            prediction=prediction[i],
                            token=tokens[i],
                            full_text=input_sample.full_text,
                            metadata=input_sample.metadata,
                        )
                    )
                elif new_annotation[i] == "O":
                    mistakes.append(
                        ModelError(
                            error_type="FP",
                            annotation=new_annotation[i],
                            prediction=prediction[i],
                            token=tokens[i],
                            full_text=input_sample.full_text,
                            metadata=input_sample.metadata,
                        )
                    )
                else:
                    mistakes.append(
                        ModelError(
                            error_type="Wrong entity",
                            annotation=new_annotation[i],
                            prediction=prediction[i],
                            token=tokens[i],
                            full_text=input_sample.full_text,
                            metadata=input_sample.metadata,
                        )
                    )

        return results, mistakes

    def _adjust_per_entities(self, tags):
        if self.entities_to_keep:
            return [tag if tag in self.entities_to_keep else "O" for tag in tags]
        else:
            return tags

    @staticmethod
    def _to_io(tags):
        """
        Translates BILUO/BIO/IOB to IO - only In or Out of entity.
        ['B-PERSON','I-PERSON','L-PERSON'] is translated into
        ['PERSON','PERSON','PERSON']
        :param tags: the input tags in BILUO/IOB/BIO format
        :return: a new list of IO tags
        """
        return [tag[2:] if "-" in tag else tag for tag in tags]

    def evaluate_sample(
        self, sample: InputSample, prediction: List[str]
    ) -> EvaluationResult:
        if self.verbose:
            print("Input sentence: {}".format(sample.full_text))

        results, mistakes = self.compare(input_sample=sample, prediction=prediction)
        return EvaluationResult(results, mistakes, sample.full_text)

    def evaluate_all(self, dataset: List[InputSample]) -> List[EvaluationResult]:
        evaluation_results = []
        if self.model.entity_mapping:
            print(
                f"Mapping entity values using this dictionary: {self.model.entity_mapping}"
            )
        for sample in tqdm(dataset, desc=f"Evaluating {self.model.__class__}"):

            # Align tag values to the ones expected by the model
            self.model.align_entity_types(sample)

            # Predict
            prediction = self.model.predict(sample)

            # Remove entities not requested
            prediction = self.model.filter_tags_in_supported_entities(prediction)

            # Switch to requested labeling scheme (IO/BIO/BILUO)
            prediction = self.model.to_scheme(prediction)

            evaluation_result = self.evaluate_sample(
                sample=sample, prediction=prediction
            )
            evaluation_results.append(evaluation_result)

        return evaluation_results

    @staticmethod
    def align_entity_types(
        input_samples: List[InputSample],
        entities_mapping: Dict[str, str] = None,
        allow_missing_mappings: bool = False,
    ) -> List[InputSample]:
        """
        Change input samples to conform with Presidio's entities
        :return: new list of InputSample
        """

        new_input_samples = input_samples.copy()

        # A list that will contain updated input samples,
        new_list = []

        for input_sample in new_input_samples:
            contains_field_in_mapping = False
            new_spans = []
            # Update spans to match the entity types in the values of entities_mapping
            for span in input_sample.spans:
                if span.entity_type in entities_mapping.keys():
                    new_name = entities_mapping.get(span.entity_type)
                    span.entity_type = new_name
                    contains_field_in_mapping = True

                    new_spans.append(span)
                else:
                    if not allow_missing_mappings:
                        raise ValueError(
                            f"Key {span.entity_type} cannot be found in the provided entities_mapping"
                        )
            input_sample.spans = new_spans

            # Update tags in case this sample has relevant entities for evaluation
            if contains_field_in_mapping:
                for i, tag in enumerate(input_sample.tags):
                    has_prefix = "-" in tag
                    if has_prefix:
                        prefix = tag[:2]
                        clean = tag[2:]
                    else:
                        prefix = ""
                        clean = tag

                    if clean in entities_mapping.keys():
                        new_name = entities_mapping.get(clean)
                        input_sample.tags[i] = "{}{}".format(prefix, new_name)
                    else:
                        input_sample.tags[i] = "O"

            new_list.append(input_sample)

        return new_list
        # Iterate on all samples

    def calculate_score(
        self,
        evaluation_results: List[EvaluationResult],
        entities: Optional[List[str]] = None,
        beta: float = 2.5,
    ) -> EvaluationResult:
        """
        Returns the pii_precision, pii_recall, f_measure either and number of records for each entity
        or for all entities (ignore_entity_type = True)
        :param evaluation_results: List of EvaluationResult
        :param entities: List of entities to calculate score to. Default is None: all entities
        :param beta: F measure beta value
        between different entity types, or to treat these as misclassifications
        :return: EvaluationResult with precision, recall and f measures
        """

        # aggregate results
        all_results = sum([er.results for er in evaluation_results], Counter())

        # compute pii_recall per entity
        entity_recall = {}
        entity_precision = {}
        n = {}
        if not entities:
            entities = list(set([x[0] for x in all_results.keys() if x[0] != "O"]))

        for entity in entities:
            # all annotation of given type
            annotated = sum([all_results[x] for x in all_results if x[0] == entity])
            predicted = sum([all_results[x] for x in all_results if x[1] == entity])
            n[entity] = annotated
            tp = all_results[(entity, entity)]

            if annotated > 0:
                entity_recall[entity] = tp / annotated
            else:
                entity_recall[entity] = np.NaN

            if predicted > 0:
                per_entity_tp = all_results[(entity, entity)]
                entity_precision[entity] = per_entity_tp / predicted
            else:
                entity_precision[entity] = np.NaN

        # compute pii_precision and pii_recall
        annotated_all = sum([all_results[x] for x in all_results if x[0] != "O"])
        predicted_all = sum([all_results[x] for x in all_results if x[1] != "O"])
        if annotated_all > 0:
            pii_recall = (
                sum(
                    [
                        all_results[x]
                        for x in all_results
                        if (x[0] != "O" and x[1] != "O")
                    ]
                )
                / annotated_all
            )
        else:
            pii_recall = np.NaN
        if predicted_all > 0:
            pii_precision = (
                sum(
                    [
                        all_results[x]
                        for x in all_results
                        if (x[0] != "O" and x[1] != "O")
                    ]
                )
                / predicted_all
            )
        else:
            pii_precision = np.NaN
        # compute pii_f_beta-score
        pii_f_beta = self.f_beta(pii_precision, pii_recall, beta)

        # aggregate errors
        errors = []
        for res in evaluation_results:
            if res.model_errors:
                errors.extend(res.model_errors)

        evaluation_result = EvaluationResult(
            results=all_results,
            model_errors=errors,
            pii_precision=pii_precision,
            pii_recall=pii_recall,
            entity_recall_dict=entity_recall,
            entity_precision_dict=entity_precision,
            n_dict=n,
            pii_f=pii_f_beta,
            n=sum(n.values()),
        )

        return evaluation_result

    @staticmethod
    def precision(tp: int, fp: int) -> float:
        return tp / (tp + fp + 1e-100)

    @staticmethod
    def recall(tp: int, fn: int) -> float:
        return tp / (tp + fn + 1e-100)

    @staticmethod
    def f_beta(precision: float, recall: float, beta: float) -> float:
        """
        Returns the F score for precision, recall and a beta parameter
        :param precision: a float with the precision value
        :param recall: a float with the recall value
        :param beta: a float with the beta parameter of the F measure,
        which gives more or less weight to precision
        vs. recall
        :return: a float value of the f(beta) measure.
        """
        if np.isnan(precision) or np.isnan(recall) or (precision == 0 and recall == 0):
            return np.nan

        return ((1 + beta ** 2) * precision * recall) / (
            ((beta ** 2) * precision) + recall
        )



from .span_to_tag import span_to_tag, tokenize, io_to_scheme
from .data_objects import Span, InputSample
from .validation import (
    split_dataset,
    split_by_template,
    get_samples_by_pattern,
    group_by_template,
    save_to_json,
)

from dotenv import load_dotenv

load_dotenv()  # take environment variables from .env.


__all__ = [
    "span_to_tag",
    "tokenize",
    "io_to_scheme",
    "Span",
    "InputSample",
    "split_dataset",
    "split_by_template",
    "get_samples_by_pattern",
    "group_by_template",
    "save_to_json",
]



from typing import List, Optional, Dict

import spacy

from presidio_evaluator import InputSample
from presidio_evaluator.data_objects import PRESIDIO_SPACY_ENTITIES
from presidio_evaluator.models import BaseModel


class SpacyModel(BaseModel):
    def __init__(
        self,
        model: spacy.language.Language = None,
        model_name: str = None,
        entities_to_keep: List[str] = None,
        verbose: bool = False,
        labeling_scheme: str = "BIO",
        entity_mapping: Optional[Dict[str, str]] = PRESIDIO_SPACY_ENTITIES,
    ):
        super().__init__(
            entities_to_keep=entities_to_keep,
            verbose=verbose,
            labeling_scheme=labeling_scheme,
            entity_mapping=entity_mapping,
        )

        if model is None:
            if model_name is None:
                raise ValueError("Either model_name or model object must be supplied")
            self.model = spacy.load(model_name)
        else:
            self.model = model

    def predict(self, sample: InputSample) -> List[str]:
        """
        Predict a list of tags for an inpuit sample.
        :param sample: InputSample
        :return: list of tags
        """
        doc = self.model(sample.full_text)
        tags = self._get_tags_from_doc(doc)
        if len(doc) != len(sample.tokens):
            print("mismatch between input tokens and new tokens")

        return tags

    @staticmethod
    def _get_tags_from_doc(doc):
        tags = [token.ent_type_ if token.ent_type_ != "" else "O" for token in doc]
        return tags



import re
from typing import List, Union, Optional

from faker import Generator

from presidio_evaluator.data_generator.faker_extensions import (
    FakerSpansResult,
    FakerSpan,
)

_re_token = re.compile(r"\{\{\s*(\w+)(:\s*\w+?)?\s*\}\}")


class SpanGenerator(Generator):
    """Generator which also returns the indices of fake values.

    :example:
    >>>from faker import Faker
    >>>from presidio_evaluator.data_generator.faker_extensions import SpanGenerator

    >>>generator = SpanGenerator()
    >>>faker = Faker(generator=generator)
    >>>res = faker.parse("My child's name is {{name}}", add_spans=True)

    >>>res.spans
        [{"value": "Daniel Gallagher", "start": 19, "end": 35, "type": "name"}]
    >>>res.fake
        "My child's name is Daniel Gallagher"
    >>>str(res)
        "My child's name is Daniel Gallagher"
    """

    def parse(
        self, text: str, add_spans: bool = False, template_id: Optional[int] = None
    ) -> Union[str, FakerSpansResult]:
        """Parses a Faker template.

        This replaces the original parse method to introduce spans.
        :param text: Text holding the faker template, e.g. "My name is {{name}}".
        :param add_spans: Whether to return the spans of each fake value in the output string
        :param template_id: Template ID to be returned with the output
        """

        # Create Span objects for original placeholders
        spans = self._match_to_span(text)

        # Reverse for easier index handling while replacing
        spans = sorted(spans, reverse=True, key=lambda x: x.start)

        fake_text = ""
        prev_end = len(text)  # we are going backwards

        # Update indices and fake text based on new values
        for i, span in enumerate(spans):
            formatter = span.type
            old_len = len(formatter) + 4  # adding two curly brackets
            new_len = len(str(span.value))

            # Update full text
            fake_text = str(text[span.end : prev_end]) + str(fake_text)
            fake_text = str(span.value) + str(fake_text)
            prev_end = span.start

            if add_spans:  # skip if spans aren't required
                # Update span indices
                delta = new_len - old_len
                span.end = span.end + delta
                span.type = formatter.strip()

                # Update previously inserted spans since indices shifted
                for j in range(0, i):
                    spans[j].start += delta
                    spans[j].end += delta

        # Add the beginning of the sentence
        fake_text = text[0:prev_end] + fake_text

        return (
            FakerSpansResult(
                fake=fake_text, spans=spans, template=text, template_id=template_id
            )
            if add_spans
            else fake_text
        )

    def _match_to_span(self, text: str, **kwargs) -> List[FakerSpan]:
        matches = _re_token.finditer(text)

        results: List[FakerSpan] = []
        for match in matches:
            formatter = match.group()[2:-2]
            results.append(
                FakerSpan(
                    type=formatter,
                    start=match.start(),
                    end=match.end(),
                    value=str(self.format(formatter.strip(), **kwargs)),
                )
            )

        return results



from typing import Dict, Optional, List, Any

from faker.providers import DynamicProvider
from faker.generator import _re_token
from presidio_evaluator.data_generator.faker_extensions import FakerSpan, SpanGenerator


class RecordGenerator(SpanGenerator):
    """Prioritizes the sampling of values found in input records,
    in order to allow semantically similar elements to be drawn together.

    For example, for the input template
    "My first name is {{name}} and my email is {{email}}",
    assuming the user added a list of records containing name and email
    (e.g. `[{"name": "Stephane Smith", "email": "stephane.smith@gmail.com"}]`),
    this generator will sample the name and email from the same record
    and not independently.
    In case the template contains entities not found in the record,
    the generator will use regular Faker providers.
    In case the template contains the same type multiple times,
    the first would be taken from the record and the next ones
    from regular Faker providers.

    :example:
    >>>from faker import Faker
    >>>from presidio_evaluator.data_generator.faker_extensions import RecordGenerator

    >>>records = [
    >>>     {"name": "Alan", "email": "alan@a.com"},
    >>>     {"name": "Barry", "email": "barry@b.com"},
    >>>     {"name": "Cynthia", "email": "cynthia@c.com"},
    >>>     ]

    >>>generator = RecordGenerator(records=records)
    >>>faker = Faker(generator=generator)

    >>># Sample fake values from the same record
    >>>faker.parse("I'm {{name}} and my email is {{email}}.")

    I'm Alan and my email is alan@a.com.

    >>># Using more than one type will use regular Faker providers:
    >>>faker.parse("{{name}}, {{name}} and {{name}} will email {{email}}.")

    Cynthia, Manuel Gonzales and Jillian Riley will email cynthia@c.com.

    >>># Sample from a Pandas DataFrame
    >>>import pandas as pd
    >>>df = pd.DataFrame({"name":["a","b","c"],"email":["a@a","b@b","c@c"]}) # or read from file
    >>>records = df.to_dict(orient="records")
    >>>generator = RecordGenerator(records=records)
    >>>faker = Faker(generator=generator)
    >>>faker.parse("I'm {{name}} and my email is {{email}}")

    I'm a and my email is a@a

    >>># Return spans of fake values
    >>>res = faker.parse("I'm {{name}} and my email is {{email}}",add_spans=True)

    {"fake": "I'm c and my email is c@c",
     "spans": "[{\"value\": \"c@c\", \"start\": 22, \"end\": 25, \"type\": \"email\"},
     {\"value\": \"c\", \"start\": 4, \"end\": 5, \"type\": \"name\"}]"
     }

    """

    def __init__(self, records: Optional[List[Dict]] = None):
        super().__init__()
        self.records = records

        if self.records:
            for record in self.records:
                if not isinstance(record, Dict):
                    raise TypeError("Each element should be of type Dict")

        # Use an internal provider to sample from the input elements
        self.dynamic_record_provider = DynamicProvider(
            provider_name="", elements=records, generator=self
        )

    def _get_random_record(self):
        return self.dynamic_record_provider.get_random_value().copy()

    def _match_to_span(self, text: str, **kwargs) -> List[FakerSpan]:
        """Adds logic for sampling from input records if possible."""
        matches = _re_token.finditer(text)
        # Sample one record (Dict containing fake values)
        record = self._get_random_record()

        results: List[FakerSpan] = []
        for match in matches:
            formatter = match.group()[2:-2]
            stripped = formatter.strip()

            value = str(self.format(formatter=stripped, record=record))
            if stripped in record:
                del record[stripped]  # Remove in order not to sample twice

            results.append(
                FakerSpan(
                    type=formatter,
                    start=match.start(),
                    end=match.end(),
                    value=value,
                )
            )

        return results

    def format(self, formatter: str, *args: Any, **kwargs: Any) -> str:
        """Fill in fake data. If the input record has the requested entity, return its value."""
        record = kwargs.get("record")
        if not record or not record.get(
            formatter
        ):  # type not in record, go to default faker
            return super().format(formatter)

        return record[formatter]



import os

from .experiment_tracker import ExperimentTracker

try:
    from comet_ml import Experiment
except ImportError:
    Experiment = None


def get_experiment_tracker():
    framework = os.environ.get("tracking_framework", None)
    if not framework or not Experiment:
        return ExperimentTracker()
    elif framework == "comet":
        return Experiment(
            api_key=os.environ.get("API_KEY"),
            project_name=os.environ.get("PROJECT_NAME"),
            workspace=os.environ.get("WORKSPACE"),
        )


__all__ = ["ExperimentTracker", "get_experiment_tracker"]



from typing import List, Optional, Dict

from presidio_analyzer import AnalyzerEngine

from presidio_evaluator import InputSample, span_to_tag
from presidio_evaluator.models import BaseModel


class PresidioAnalyzerWrapper(BaseModel):
    def __init__(
        self,
        analyzer_engine: Optional[AnalyzerEngine] = None,
        entities_to_keep: List[str] = None,
        verbose: bool = False,
        labeling_scheme: str = "BIO",
        score_threshold: float = 0.4,
        language: str = "en",
        entity_mapping: Optional[Dict[str, str]] = None,
    ):
        """
        Evaluation wrapper for the Presidio Analyzer
        :param analyzer_engine: object of type AnalyzerEngine (from presidio-analyzer)
        """
        super().__init__(
            entities_to_keep=entities_to_keep,
            verbose=verbose,
            labeling_scheme=labeling_scheme,
            entity_mapping=entity_mapping,
        )
        self.score_threshold = score_threshold
        self.language = language

        if not analyzer_engine:
            analyzer_engine = AnalyzerEngine()
            self._update_recognizers_based_on_entities_to_keep(analyzer_engine)
        self.analyzer_engine = analyzer_engine

    def predict(self, sample: InputSample) -> List[str]:

        results = self.analyzer_engine.analyze(
            text=sample.full_text,
            entities=self.entities,
            language="en",
            score_threshold=self.score_threshold,
        )
        starts = []
        ends = []
        scores = []
        tags = []
        #
        for res in results:
            starts.append(res.start)
            ends.append(res.end)
            tags.append(res.entity_type)
            scores.append(res.score)

        response_tags = span_to_tag(
            scheme="IO",
            text=sample.full_text,
            starts=starts,
            ends=ends,
            tokens=sample.tokens,
            scores=scores,
            tags=tags,
        )
        return response_tags

    # Mapping between dataset entities and Presidio entities. Key: Dataset entity, Value: Presidio entity
    presidio_entities_map = {
        "PERSON": "PERSON",
        "GPE": "LOCATION",
        "EMAIL_ADDRESS": "EMAIL_ADDRESS",
        "CREDIT_CARD": "CREDIT_CARD",
        "FIRST_NAME": "PERSON",
        "LAST_NAME": "PERSON",
        "PHONE_NUMBER": "PHONE_NUMBER",
        "BIRTHDAY": "DATE_TIME",
        "DATE_TIME": "DATE_TIME",
        "DOMAIN_NAME": "DOMAIN_NAME",
        "CITY": "LOCATION",
        "ADDRESS": "LOCATION",
        "NATIONALITY": "LOCATION",
        "LOCATION": "LOCATION",
        "IBAN_CODE": "IBAN_CODE",
        "URL": "DOMAIN_NAME",
        "US_SSN": "US_SSN",
        "IP_ADDRESS": "IP_ADDRESS",
        "ORGANIZATION": "ORG",
        "US_DRIVER_LICENSE": "US_DRIVER_LICENSE",
        "NRP": "NRP",
        "TITLE": "O",  # not supported
        "PREFIX": "O",  # not supported
        "STREET_ADDRESS": "O",  # not supported
        "ZIP_CODE": "O",  # not supported
        "AGE": "O",  # not supported
        "O": "O",
    }

    def _update_recognizers_based_on_entities_to_keep(
        self, analyzer_engine: AnalyzerEngine
    ):
        """Check if there are any entities not supported by this presidio instance.
        Add ORGANIZATION as it is removed by default

        """
        supported_entities = analyzer_engine.get_supported_entities(
            language=self.language
        )
        print("Entities supported by this Presidio Analyzer instance:")
        print(", ".join(supported_entities))

        if not self.entities:
            self.entities = supported_entities

        for entity in self.entities:
            if entity not in supported_entities:
                print(
                    f"Entity {entity} is not supported by this instance of Presidio Analyzer Engine"
                )

        if "ORGANIZATION" in self.entities and "ORGANIZATION" not in supported_entities:
            recognizers = analyzer_engine.get_recognizers()
            spacy_recognizer = [
                rec
                for rec in recognizers
                if rec.name == "SpacyRecognizer" or rec.name == "StanzaRecognizer"
            ]
            if len(spacy_recognizer):
                spacy_recognizer = spacy_recognizer[0]
                spacy_recognizer.supported_entities.append("ORGANIZATION")
                self.entities.append("ORGANIZATION")
                print("Added ORGANIZATION as a supported entity from spaCy/Stanza")



from typing import List, Dict

import spacy

from presidio_evaluator.data_objects import PRESIDIO_SPACY_ENTITIES

try:
    from flair.data import Sentence
    from flair.models import SequenceTagger
    from flair.tokenization import SpacyTokenizer
except ImportError:
    print("Flair is not installed by default")

from presidio_evaluator import InputSample, tokenize, span_to_tag
from presidio_evaluator.models import BaseModel


class FlairModel(BaseModel):
    """
    Evaluator for Flair models
    :param model: model of type SequenceTagger
    :param model_path:
    :param entities_to_keep:
    :param verbose:
    and model expected entity types
    """

    def __init__(
        self,
        model=None,
        model_path: str = None,
        entities_to_keep: List[str] = None,
        verbose: bool = False,
        entity_mapping: Dict[str, str] = PRESIDIO_SPACY_ENTITIES,
    ):

        super().__init__(
            entities_to_keep=entities_to_keep,
            verbose=verbose,
            entity_mapping=entity_mapping,
        )
        if model is None:
            if model_path is None:
                raise ValueError("Either model_path or model object must be supplied")
            self.model = SequenceTagger.load(model_path)
        else:
            self.model = model

        self.spacy_tokenizer = SpacyTokenizer(model=spacy.load("en_core_web_sm"))

    def predict(self, sample: InputSample) -> List[str]:

        sentence = Sentence(text=sample.full_text, use_tokenizer=self.spacy_tokenizer)
        self.model.predict(sentence)

        ents = sentence.get_spans("ner")
        if ents:
            tags, texts, start, end = zip(
                *[(ent.tag, ent.text, ent.start_pos, ent.end_pos) for ent in ents]
            )

            tags = [
                tag if tag != "PER" else "PERSON" for tag in tags
            ]  # Flair's tag for PERSON is PER

            # Flair tokens might not be consistent with spaCy's tokens (even when using spacy tokenizer)
            # Use spacy tokenization and not stanza to maintain consistency with other models:
            if not sample.tokens:
                sample.tokens = tokenize(sample.full_text)

            # Create tags (label per token) based on stanza spans and spacy tokens
            tags = span_to_tag(
                scheme="IO",
                text=sample.full_text,
                starts=start,
                ends=end,
                tags=tags,
                tokens=sample.tokens,
            )
        else:
            tags = ["O" for _ in range(len(sample.tokens))]

        if len(tags) != len(sample.tokens):
            print("mismatch between input tokens and new tokens")

        return tags



import random
from collections import OrderedDict
from pathlib import Path
from typing import Union

import pandas as pd
from faker.providers import BaseProvider
from faker.providers.address.en_US import Provider as AddressProvider
from faker.providers.phone_number.en_US import Provider as PhoneNumberProvider


class NationalityProvider(BaseProvider):
    def __init__(self, generator, nationality_file: Union[str, Path] = None):
        super().__init__(generator=generator)
        if not nationality_file:
            nationality_file = Path(
                Path(__file__).parent.parent, "raw_data", "nationalities.csv"
            ).resolve()

        self.nationality_file = nationality_file
        self.nationalities = self.load_nationalities()

    def load_nationalities(self):
        return pd.read_csv(self.nationality_file)

    def country(self):
        self.random_element(self.nationalities["country"].tolist())

    def nationality(self):
        return self.random_element(self.nationalities["nationality"].tolist())

    def nation_man(self):
        return self.random_element(self.nationalities["man"].tolist())

    def nation_woman(self):
        return self.random_element(self.nationalities["woman"].tolist())

    def nation_plural(self):
        return self.random_element(self.nationalities["plural"].tolist())


class OrganizationProvider(BaseProvider):
    def __init__(
        self,
        generator,
        organizations_file: Union[str, Path] = None,
    ):
        super().__init__(generator=generator)
        if not organizations_file:
            organizations_file = Path(
                Path(__file__).parent.parent, "raw_data", "organizations.csv"
            ).resolve()
        self.organizations_file = organizations_file
        self.organizations = self.load_organizations()

    def load_organizations(self):
        return pd.read_csv(self.organizations_file, delimiter="\t")

    def organization(self):
        return self.random_element(self.organizations["organization"].tolist())

    def company(self):
        return self.organization()


class UsDriverLicenseProvider(BaseProvider):
    def __init__(
        self,
        generator,
        us_driver_license_file: Union[str, Path] = None,
    ):
        super().__init__(generator=generator)
        if not us_driver_license_file:
            us_driver_license_file = Path(
                Path(__file__).parent.parent, "raw_data", "us_driver_licenses.csv"
            ).resolve()
        self.us_driver_license_file = us_driver_license_file
        self.us_driver_licenses = self.load_us_driver_licenses()

    def us_driver_license(self):
        return self.random_element(
            self.us_driver_licenses["us_driver_license"].tolist()
        )

    def load_us_driver_licenses(self):
        return pd.read_csv(self.us_driver_license_file, delimiter="\t")


class IpAddressProvider(BaseProvider):
    """Generating both v4 and v6 IP addresses."""

    def ip_address(self):
        if random.random() < 0.8:
            return self.generator.ipv4()
        else:
            return self.generator.ipv6()


class AgeProvider(BaseProvider):

    formats = OrderedDict(
        [
            ("%#", 0.8),
            ("%", 0.1),
            ("1.%", 0.02),
            ("2.%", 0.02),
            ("100", 0.02),
            ("101", 0.01),
            ("104", 0.01),
            ("0.%", 0.02),
        ]
    )

    def age(self):
        return self.numerify(
            self.random_elements(elements=self.formats, length=1, use_weighting=True)[0]
        )


class AddressProviderNew(AddressProvider):
    """
    Extending the Faker AddressProvider with additional templates
    """

    address_formats = OrderedDict(
        (
            (
                "{{building_number}} {{street_name}} {{secondary_address}} {{city}} {{state}}",
                5.0,
            ),
            (
                "{{building_number}} {{street_name}} {{secondary_address}} {{city}} {{state_abbr}}",
                5.0,
            ),
            (
                "{{building_number}} {{street_name}} {{secondary_address}} {{city}} {{country}}",
                5.0,
            ),
            (
                "{{building_number}} {{street_name}}\n {{secondary_address}}\n {{city}}\n {{country}}",
                5.0,
            ),
            (
                "{{building_number}} {{street_name}}\n {{secondary_address}}\n {{city}}\n {{country}} {{postcode}}",
                5.0,
            ),
            (
                "{{street_name}} {{street_name}}\n {{secondary_address}}\n {{city}}\n {{country}} {{postcode}}",
                5.0,
            ),
            ("the corner of {{street_name}} and {{street_name}}", 3.0),
            ("{{first_name}} and {{street_name}}", 3.0),
            ("{{street_address}}, {{city}}, {{country}}", 5.0),
            (
                "{{street_address}} {{secondary_address}}, {{city}}, {{country}} {{postcode}}",
                5.0,
            ),
            ("{{street_address}}\n{{city}}, {{state_abbr}} {{postcode}}", 25.0),
            ("{{street_address}}\n{{city}}\n, {{state_abbr}}\n {{postcode}}", 25.0),
            (
                "{{street_address}}\n{{city}}\n, {{state_abbr}}\n {{country}} {{postcode}}",
                25.0,
            ),
            #  military address formatting.
            ("{{military_apo}}\nAPO {{military_state}} {{postcode}}", 1.0),
            (
                "{{military_ship}} {{last_name}}\nFPO {{military_state}} {{postcode}}",
                1.0,
            ),
            ("{{military_dpo}}\nDPO {{military_state}} {{postcode}}", 1.0),
        )
    )


class PhoneNumberProviderNew(PhoneNumberProvider):
    """
    Similar to the default PhoneNumberProvider, with different formats
    """

    formats = (
        # US
        "##########",
        "##########",
        "###-###-####",
        "###-###-####",
        "###-#######",
        # UK
        "07700 ### ###",
        "07700 ######",
        "07700######",
        "(07700) ### ###",
        "(07700) ######",
        "(07700)######",
        "+447700 ### ###",
        "+447700 ######",
        "+447700######",
        # India
        "+91##########",
        "0##########",
        "##########",
        # Switzerland
        "+41 2# ### ## ##",
        "+41 3# ### ## ##",
        "+41 4# ### ## ##",
        "+41 5# ### ## ##",
        "+41 6# ### ## ##",
        "+41 7# ### ## ##",
        "+41 8# ### ## ##",
        "+41 9# ### ## ##",
        "+41 (0)2# ### ## ##",
        "+41 (0)3# ### ## ##",
        "+41 (0)4# ### ## ##",
        "+41 (0)5# ### ## ##",
        "+41 (0)6# ### ## ##",
        "+41 (0)7# ### ## ##",
        "+41 (0)8# ### ## ##",
        "+41 (0)9# ### ## ##",
        "+46 (0)8 ### ### ##",
        "+46 (0)## ## ## ##",
        "+46 (0)### ### ##",
        # Optional 10-digit local phone number format
        "(###)###-####",
        "(###)###-####",
        "(###)###-####",
        "(###)###-####",
        # Non-standard 10-digit phone number format
        "###.###.####",
        "###.###.####",
        # Standard 10-digit phone number format with extensions
        "###-###-####x###",
        "###-###-####x####",
        # Optional 10-digit local phone number format with extensions
        "(###)###-####x###",
        "(###)###-####x####",
        # Non-standard 10-digit phone number format with extensions
        "###.###.####x###",
        "###.###.####x####",
        # Standard 11-digit phone number format
        "+1-###-###-####",
        "001-###-###-####",
        # Standard 11-digit phone number format with extensions
        "+1-###-###-####x###",
    )



from typing import List

import pandas as pd

try:
    from flair.data import Corpus, Sentence
    from flair.datasets import ColumnCorpus
    from flair.embeddings import (
        TokenEmbeddings,
        WordEmbeddings,
        StackedEmbeddings,
        FlairEmbeddings,
        BertEmbeddings,
    )
    from flair.models import SequenceTagger
    from flair.trainers import ModelTrainer
except ImportError:
    print("Flair is not installed")

from presidio_evaluator import InputSample

from os import path


class FlairTrainer:
    """
    Helper class for training Flair models
    """

    @staticmethod
    def to_flair_row(text: str, pos: str, label: str) -> str:
        """
        Turn text, part of speech and label into one row.
        :return: str
        """
        return "{} {} {}".format(text, pos, label)

    def to_flair(self, df: pd.DataFrame, outfile: str = "flair_train.txt") -> None:
        """Translate a pd.DataFrame to a flair dataset."""
        sentence = 0
        flair = []
        for row in df.itertuples():
            if row.sentence != sentence:
                sentence += 1
                flair.append("")
            else:
                flair.append(self.to_flair_row(row.text, row.pos, row.label))

        if outfile:
            with open(outfile, "w", encoding="utf-8") as f:
                for item in flair:
                    f.write("{}\n".format(item))

    def create_flair_corpus(
        self, train_samples_path, test_samples_path, val_samples_path
    ):
        """
        Create a flair Corpus object and saive it to train, test, validation files.
        :param train_samples_path: Path to train samples
        :param test_samples_path: Path to test samples
        :param val_samples_path: Path to validation samples
        :return:
        """
        if not path.exists("flair_train.txt"):
            train_samples = InputSample.read_dataset_json(train_samples_path)
            train_tagged = [sample for sample in train_samples if len(sample.spans) > 0]
            print(
                f"Kept {len(train_tagged)} train samples after removal of non-tagged samples"
            )
            train_data = InputSample.create_conll_dataset(train_tagged)
            self.to_flair(train_data, outfile="flair_train.txt")

        if not path.exists("flair_test.txt"):
            test_samples = InputSample.read_dataset_json(test_samples_path)
            test_data = InputSample.create_conll_dataset(test_samples)
            self.to_flair(test_data, outfile="flair_test.txt")

        if not path.exists("flair_val.txt"):
            val_samples = InputSample.read_dataset_json(val_samples_path)
            val_data = InputSample.create_conll_dataset(val_samples)
            self.to_flair(val_data, outfile="flair_val.txt")

    @staticmethod
    def read_corpus(data_folder: str):
        """
        Read Flair Corpus object.
        :param data_folder: Path with files
        :return: Corpus object
        """
        columns = {0: "text", 1: "pos", 2: "ner"}
        corpus = ColumnCorpus(
            data_folder,
            columns,
            train_file="flair_train.txt",
            test_file="flair_val.txt",
            dev_file="flair_test.txt",
        )
        return corpus

    @staticmethod
    def train(corpus):
        """
        Train a Flair model
        :param corpus: Corpus object
        :return:
        """
        print(corpus)

        # 2. what tag do we want to predict?
        tag_type = "ner"

        # 3. make the tag dictionary from the corpus
        tag_dictionary = corpus.make_tag_dictionary(tag_type=tag_type)
        print(tag_dictionary.idx2item)

        # 4. initialize embeddings
        embedding_types: List[TokenEmbeddings] = [
            WordEmbeddings("glove"),
            FlairEmbeddings("news-forward"),
            FlairEmbeddings("news-backward"),
        ]

        embeddings: StackedEmbeddings = StackedEmbeddings(embeddings=embedding_types)

        # 5. initialize sequence tagger

        tagger: SequenceTagger = SequenceTagger(
            hidden_size=256,
            embeddings=embeddings,
            tag_dictionary=tag_dictionary,
            tag_type=tag_type,
            use_crf=True,
        )

        # 6. initialize trainer

        trainer: ModelTrainer = ModelTrainer(tagger, corpus)

        checkpoint = "resources/taggers/presidio-ner/checkpoint.pt"
        # trainer = ModelTrainer.load_checkpoint(checkpoint, corpus)
        trainer.train(
            "resources/taggers/presidio-ner",
            learning_rate=0.1,
            mini_batch_size=32,
            max_epochs=150,
            checkpoint=True,
        )

        sentence = Sentence("I am from Jerusalem")
        # run NER over sentence
        tagger.predict(sentence)

        print(sentence)
        print("The following NER tags are found:")

        # iterate over entities and print
        for entity in sentence.get_spans("ner"):
            print(entity)


if __name__ == "__main__":
    train_samples = "../../data/train_Dec-19-2021.json"
    test_samples = "../../data/test_Dec-19-2021.json"
    val_samples = "../../data/validation_Dec-19-2021.json"

    trainer = FlairTrainer()
    trainer.create_flair_corpus(train_samples, test_samples, val_samples)

    corpus = trainer.read_corpus("")
    trainer.train(corpus)



import json
import time
from typing import Dict, List


class ExperimentTracker:
    def __init__(self):
        self.parameters = dict()
        self.metrics = dict()
        self.dataset_info = None
        self.confusion_matrix = None
        self.labels = None

    def log_parameter(self, key: str, value: object):
        self.parameters[key] = value

    def log_parameters(self, parameters: Dict):
        for k, v in parameters.items():
            self.log_parameter(k, v)

    def log_metric(self, key: str, value: object):
        self.metrics[key] = value

    def log_metrics(self, metrics: Dict):
        for k, v in metrics.items():
            self.log_metric(k, v)

    def log_dataset_hash(self, data: str):
        pass

    def log_dataset_info(self, name: str):
        self.dataset_info = name

    def __str__(self):
        return json.dumps(self.__dict__)

    def log_confusion_matrix(
        self,
        matrix: List[List[int]],
        labels=List[str],
    ):
        self.confusion_matrix = matrix
        self.labels = labels

    def start(self):
        pass

    def end(self):
        datetime_val = time.strftime("%Y%m%d-%H%M%S")
        filename = f"experiment_{datetime_val}.json"
        print(f"saving experiment data to {filename}")
        with open(filename, 'w') as json_file:
            json.dump(self.__dict__, json_file)



from typing import List, Optional

import spacy
from spacy.tokens import Doc

loaded_spacy = {}


def get_spacy(loaded_spacy=loaded_spacy, model_version="en_core_web_sm"):
    if model_version not in loaded_spacy:
        print("loading model {}".format(model_version))
        loaded_spacy[model_version] = spacy.load(model_version)
    return loaded_spacy[model_version]


def tokenize(text, model_version="en_core_web_sm") -> Doc:
    return get_spacy(model_version=model_version)(text)


def _get_detailed_tags_for_span(scheme: str, cur_tags: List[str]) -> List[str]:
    """
    Replace IO tags (e.g. O PERSON PERSON) with BIO/BILUO tags.
    """

    if all([tag == "O" for tag in cur_tags]):
        return cur_tags

    return_tags = []
    if len(cur_tags) == 1:
        if scheme == "BILUO":
            return_tags.append(f"U-{cur_tags[0]}")
        else:
            return_tags.append(f"B-{cur_tags[0]}")
    elif len(cur_tags) > 0:
        tg = cur_tags[0]
        for j in range(0, len(cur_tags)):
            if j == 0:
                return_tags.append(f"B-{tg}")
            elif j == len(cur_tags) - 1:
                if scheme == "BILUO":
                    return_tags.append(f"L-{tg}")
                else:
                    return_tags.append(f"I-{tg}")
            else:
                return_tags.append(f"I-{tg}")
    return return_tags


def _sort_spans(start, end, tag, score):
    if len(start) > 0:
        tpl = [
            (a, b, c, d)
            for a, b, c, d in sorted(
                zip(start, end, tag, score), key=lambda pair: pair[0]
            )
        ]
        start, end, tag, score = [[x[i] for x in tpl] for i in range(len(tpl[0]))]
    return start, end, tag, score


def _handle_overlaps(start, end, tag, score):
    start, end, tag, score = _sort_spans(start, end, tag, score)
    if len(start) == 0:
        return start, end, tag, score
    max_end = max(end)
    index = min(start)
    number_of_spans = len(start)
    i = 0
    while i < number_of_spans - 1:
        for j in range(i + 1, number_of_spans):
            # Span j intersects with span i
            if start[i] <= start[j] <= end[i]:
                # i's score is higher, remove intersecting part
                if score[i] > score[j]:
                    # j is contained within i but has lower score, remove
                    if start[i] <= end[j] <= end[i]:
                        score[j] = 0
                    # else, j continues after i ended:
                    else:
                        start[j] = end[i] + 1
                # j's score is higher, break i
                else:
                    # If i finishes after j ended, split i
                    if end[j] < end[i]:
                        # create new span at the end
                        start.append(end[j] + 1)
                        end.append(end[i])
                        score.append(score[i])
                        tag.append(tag[i])
                        number_of_spans += 1
                        # truncate the current i to end at start(j)
                        end[i] = start[j] - 1
                    # else, i finishes before j ended. truncate i
                    else:
                        end[i] = start[j] - 1

        i += 1
    start, end, tag, score = _sort_spans(start, end, tag, score)
    return start, end, tag, score


def span_to_tag(
    scheme: str,
    text: str,
    starts: List[int],
    ends: List[int],
    tags: List[str],
    scores: Optional[List[float]] = None,
    tokens: Optional[Doc] = None,
    token_model_version: str = "en_core_web_sm"
) -> List[str]:
    """
    Turns a list of start and end values with corresponding labels, into a NER
    tagging (BILUO,BIO/IOB)
    :param scheme: labeling scheme, either BILUO, BIO/IOB or IO
    :param text: input text
    :param tokens: text tokenized to tokens
    :param starts: list of indices where entities in the text start
    :param ends: list of indices where entities in the text end
    :param tags: list of entity names
    :param scores: score of tag (confidence)
    :param token_model_version: version of the model used for tokenization if no tokens provided
    :return: list of strings, representing either BILUO or BIO for the input
    """

    if not scores:
        # assume all scores are of equal weight
        scores = [0.5 for start in starts]

    starts, ends, tags, scores = _handle_overlaps(starts, ends, tags, scores)

    if not tokens:
        tokens = tokenize(text, token_model_version)

    io_tags = []
    for token in tokens:
        found = False
        for span_index in range(0, len(starts)):
            span_start_in_token = (
                token.idx <= starts[span_index] <= token.idx + len(token.text)
            )
            span_end_in_token = (
                token.idx <= ends[span_index] <= token.idx + len(token.text)
            )
            if (
                starts[span_index] <= token.idx < ends[span_index]
            ):  # token start is between start and end
                io_tags.append(tags[span_index])
                found = True
            elif (
                span_start_in_token and span_end_in_token
            ):  # span is within token boundaries (special case)
                io_tags.append(tags[span_index])
                found = True
            if found:
                break

        if not found:
            io_tags.append("O")

    if scheme == "IO":
        return io_tags
    else:
        return io_to_scheme(io_tags, scheme)


def io_to_scheme(io_tags: List[str], scheme: str) -> List[str]:
    """Set tagging based on scheme (BIO or BILUO).
    :param io_tags: List of tags in IO (e.g. O O O PERSON PERSON O)
    :param scheme: Requested scheme (IO, BILUO or BIO)
    """

    if scheme == "IO":
        return io_tags

    if scheme == "BILOU":
        scheme = "BILUO"


    current_tag = ""
    span_index = 0
    changes = []
    for io_tag in io_tags:
        if io_tag != current_tag:
            changes.append(span_index)
        span_index += 1
        current_tag = io_tag
    changes.append(len(io_tags))
    new_return_tags = []
    for i in range(len(changes) - 1):
        new_return_tags.extend(
            _get_detailed_tags_for_span(
                scheme=scheme, cur_tags=io_tags[changes[i] : changes[i + 1]]
            )
        )
    return new_return_tags



from typing import Dict, List

import pandas as pd
from spacy.tokens import Token


class ModelError:
    def __init__(
        self,
        error_type: str,
        annotation: str,
        prediction: str,
        token: Token,
        full_text: str,
        metadata: Dict,
    ):
        """
        Holds information about an error a model made for analysis purposes
        :param error_type: str, e.g. FP, FN, Person->Address etc.
        :param annotation: ground truth value
        :param prediction: predicted value
        :param token: token in question
        :param full_text: full input text
        :param metadata: metadata on text from InputSample
        """

        self.error_type = error_type
        self.annotation = annotation
        self.prediction = prediction
        self.token = token
        self.full_text = full_text
        self.metadata = metadata

    def __str__(self):
        return (
            "type: {}, "
            "Annotation = {}, "
            "prediction = {}, "
            "Token = {}, "
            "Full text = {}, "
            "Metadata = {}".format(
                self.error_type,
                self.annotation,
                self.prediction,
                self.token,
                self.full_text,
                self.metadata,
            )
        )

    def __repr__(self):
        return f"<ModelError {self.__str__()}"

    @staticmethod
    def most_common_fp_tokens(errors=List["ModelError"], n: int = 10, entity=None):
        """
        Print the n most common false positive tokens
        (tokens thought to be an entity)
        """
        fps = ModelError.get_false_positives(errors, entity)

        tokens = [err.token.text for err in fps]
        from collections import Counter

        by_frequency = Counter(tokens)
        most_common = by_frequency.most_common(n)
        print("Most common false positive tokens:")
        print(most_common)
        print("Example sentence with each FP token:")
        for tok, val in most_common:
            with_tok = [err for err in fps if err.token.text == tok]
            print(with_tok[0].full_text)

    @staticmethod
    def most_common_fn_tokens(errors=List["ModelError"], n: int = 10, entity=None):
        """
        Print all tokens that were missed by the model,
        including an example of the full text in which they appear.
        """
        fns = ModelError.get_false_negatives(errors, entity)

        fns_tokens = [err.token.text for err in fns]
        from collections import Counter

        by_frequency_fns = Counter(fns_tokens)
        most_common_fns = by_frequency_fns.most_common(n)
        print(most_common_fns)
        for tok, val in most_common_fns:
            with_tok = [err for err in fns if err.token.text == tok]
            print(
                "Token: {}, Annotation: {}, Full text: {}".format(
                    with_tok[0].token, with_tok[0].annotation, with_tok[0].full_text
                )
            )

    @staticmethod
    def get_errors_df(
        errors=List["ModelError"], entity: List[str] = None, error_type: str = "FN"
    ):
        """
        Get ModelErrors as pd.DataFrame
        """
        if error_type == "FN":
            filtered_errors = ModelError.get_false_negatives(errors, entity)
        elif error_type == "FP":
            filtered_errors = ModelError.get_false_positives(errors, entity)
        else:
            raise ValueError("error_type should be either FP or FN")

        if len(filtered_errors) == 0:
            print(
                "No errors of type {} and entity {} were found".format(
                    error_type, entity
                )
            )
            return None

        errors_df = pd.DataFrame.from_records(
            [error.__dict__ for error in filtered_errors]
        )
        metadata_df = pd.DataFrame(errors_df["metadata"].tolist())
        errors_df.drop(["metadata"], axis=1, inplace=True)
        new_errors_df = pd.concat([errors_df, metadata_df], axis=1)
        return new_errors_df

    @staticmethod
    def get_fps_dataframe(errors=List["ModelError"], entity: List[str] = None):
        """
        Get false positive ModelErrors as pd.DataFrame
        """
        return ModelError.get_errors_df(errors, entity, error_type="FP")

    @staticmethod
    def get_fns_dataframe(errors=List["ModelError"], entity: List[str] = None):
        """
        Get false negative ModelErrors as pd.DataFrame
        """
        return ModelError.get_errors_df(errors, entity, error_type="FN")

    @staticmethod
    def get_false_positives(errors=List["ModelError"], entity=None):
        """
        Get a list of all false positive errors in the results
        """
        if isinstance(entity, str):
            entity = [entity]

        if entity:
            return [
                model_error
                for model_error in errors
                if model_error.error_type == "FP" and model_error.prediction in entity
            ]
        else:
            return [
                model_error for model_error in errors if model_error.error_type == "FP"
            ]

    @staticmethod
    def get_false_negatives(errors=List["ModelError"], entity=None):
        """
        Get a list of all false positive negative errors in the results (False negatives and wrong entity detection)
        """
        if isinstance(entity, str):
            entity = [entity]
        if entity:
            return [
                model_error
                for model_error in errors
                if model_error.error_type != "FP" and model_error.annotation in entity
            ]
        else:
            return [
                model_error for model_error in errors if model_error.error_type != "FP"
            ]



import pickle
from typing import List, Dict, Optional

try:
    import sklearn_crfsuite
except ImportError:
    sklearn_crfsuite = None

from presidio_evaluator import InputSample
from presidio_evaluator.models import BaseModel


class CRFModel(BaseModel):
    """
    Wrapper for a CRF model.
    :param model_pickle_path: Path to pickled trained model
    """

    def __init__(
        self,
        model_pickle_path: Optional[str] = None,
        entities_to_keep: List[str] = None,
        verbose: bool = False,
        entity_mapping: Dict[str, str] = None,
    ):
        super().__init__(
            entities_to_keep=entities_to_keep,
            verbose=verbose,
            entity_mapping=entity_mapping,
        )

        if model_pickle_path:
            with open(model_pickle_path, "rb") as f:
                self.model = pickle.load(f)
        else:
            self.model = None

    def fit(
        self,
        train_samples: List[InputSample],
        algorithm="lbfgs",
        c1=0.1,
        c2=0.1,
        max_iterations=100,
        all_possible_transitions=True,
        **kwargs
    ):
        """
        Trains a simple CRF model
        :param train_samples: Training samples to train with
        :param algorithm: see `sklearn_crfsuite.CRF`
        :param c1: see `sklearn_crfsuite.CRF`
        :param c2: see `sklearn_crfsuite.CRF`
        :param max_iterations: see `sklearn_crfsuite.CRF`
        :param all_possible_transitions: see `sklearn_crfsuite.CRF`
        :return:
        """

        if not sklearn_crfsuite:
            raise ValueError("sklearn_crfsuite not installed")

        # Ignore entities not requested
        train_samples_filtered = self._ignore_unwanted_entities(train_samples)

        X_train, y_train = self._to_feature_set(train_samples_filtered)

        self.model = sklearn_crfsuite.CRF(
            algorithm=algorithm,
            c1=c1,
            c2=c2,
            max_iterations=max_iterations,
            all_possible_transitions=all_possible_transitions,
            **kwargs
        )
        self.model.fit(X_train, y_train)

    def _to_feature_set(self, dataset: List[InputSample]):

        samples_conll = InputSample.create_conll_dataset(dataset)
        sentences = samples_conll.groupby("sentence")[["text", "pos", "label"]].apply(
            lambda x: x.values.tolist()
        )

        X_train = [self.sent2features(s) for s in sentences]
        y_train = [self.sent2labels(s) for s in sentences]
        return X_train, y_train

    def predict(self, sample: InputSample) -> List[str]:
        tags = CRFModel.crf_predict(sample, self.model)

        if len(tags) != len(sample.tokens):
            print("mismatch between previous tokens and new tokens")
        return tags

    @staticmethod
    def crf_predict(sample, model):
        sample.translate_input_sample_tags()

        conll = sample.to_conll(translate_tags=True)
        sentence = [(di["text"], di["pos"], di["label"]) for di in conll]
        features = CRFModel.sent2features(sentence)
        return model.predict([features])[0]

    @staticmethod
    def word2features(sent, i):
        word = sent[i][0]
        postag = sent[i][1]

        features = {
            "bias": 1.0,
            "word.lower()": word.lower(),
            "word[-3:]": word[-3:],
            "word[-2:]": word[-2:],
            "word.isupper()": word.isupper(),
            "word.istitle()": word.istitle(),
            "word.isdigit()": word.isdigit(),
            "postag": postag,
            "postag[:2]": postag[:2],
        }
        if i > 0:
            word1 = sent[i - 1][0]
            postag1 = sent[i - 1][1]
            features.update(
                {
                    "-1:word.lower()": word1.lower(),
                    "-1:word.istitle()": word1.istitle(),
                    "-1:word.isupper()": word1.isupper(),
                    "-1:postag": postag1,
                    "-1:postag[:2]": postag1[:2],
                }
            )
        else:
            features["BOS"] = True

        if i < len(sent) - 1:
            word1 = sent[i + 1][0]
            postag1 = sent[i + 1][1]
            features.update(
                {
                    "+1:word.lower()": word1.lower(),
                    "+1:word.istitle()": word1.istitle(),
                    "+1:word.isupper()": word1.isupper(),
                    "+1:postag": postag1,
                    "+1:postag[:2]": postag1[:2],
                }
            )
        else:
            features["EOS"] = True

        return features

    @staticmethod
    def sent2features(sent):
        return [CRFModel.word2features(sent, i) for i in range(len(sent))]

    @staticmethod
    def sent2labels(sent):
        return [label for token, postag, label in sent]

    @staticmethod
    def sent2tokens(sent):
        return [token for token, postag, label in sent]
